#include "common.h"
#include "IReadFile.h"
#include "CPMDMeshFileLoader.h"

CPMDMeshFileLoader::CPMDMeshFileLoader(ISceneManager *smgr, IVideoDriver *video)
{
    SceneManager = smgr;
    VideoDriver = video;
    init();
}

void CPMDMeshFileLoader::init()
{
    // 
    pmd = NULL;
    vmd = NULL;
    AnimatedMesh = NULL;
    materialCount = 0;
}

CPMDMeshFileLoader::~CPMDMeshFileLoader(void)
{
    release();
}

void CPMDMeshFileLoader::release()
{
    // I
    if (pmd != NULL)
    {
        delete pmd;
        pmd = NULL;
    }
    if (vmd != NULL)
    {
        delete vmd;
        vmd = NULL;
    }

    if (AnimatedMesh != NULL)
    {
        AnimatedMesh->drop();
        AnimatedMesh = 0;
    }
}

bool CPMDMeshFileLoader::load(io::IReadFile *file)
{
    // t@C\
    char filename[256], path[256];
    ::GetCurrentDirectoryA(MAX_PATH, path);
    sprintf_s(filename, sizeof(filename), "%s", file->getFileName());
    printf("filename=%s\n", filename);

    int i, ret;

    if (pmd != NULL)
    {
        return false;
    }

    // PMDf[^̓ǂݍ
    pmd = new clsPMDFile;
    if ( !pmd->Open(file->getFileName()))
    {
        printf("Can't open PMD file! %s\n",
            file->getFileName());
        getchar();
        return false;
    }

    // ܂̓WCgf[^쐬
    printf("Create Joint Data\n");
    createJointData();

    // ̓}eAƃbV쐬
    materialCount = pmd->GetMaterialChunkSize();
    printf("Create Material Data\n");
    createMaterialData();

    return true;
}

bool CPMDMeshFileLoader::isALoadableFileExtension(const irr::c8 *fileName) const
{
    std::string temp(fileName);
    if (temp.find(".pmd") == std::string::npos) return false;
    else return true;
}

IAnimatedMesh* CPMDMeshFileLoader::createMesh(irr::io::IReadFile *file)
{
    if (!file)
    {
        return 0;
    }

    // XLbV쐬
    AnimatedMesh = new CSkinnedMesh();

    // ǂݍ
    if (load(file))
    {
        // evZ
        AnimatedMesh->finalize();
    }
    else
    {
        // sI
        AnimatedMesh->drop();
        AnimatedMesh = 0;
    }

    return AnimatedMesh;
}

void CPMDMeshFileLoader::createJointData()
{
    // WCg쐬
    vector3df jointPos;
    CSkinnedMesh::SJoint *parent, *joint;
    CSkinnedMesh::SPositionKey *keyP;
    CSkinnedMesh::SRotationKey *keyR;
    CSkinnedMesh::SScaleKey *keyS;

    for (int i=0; i<pmd->GetBoneChunkSize(); i++)
    {
        // em[h邩ǂ
        if (pmd->GetBoneChunk()[i].parent == (WORD)-1)
        {
            joint = AnimatedMesh->createJoint(NULL);
            jointPos.X = pmd->GetBoneChunk()[i].pos[0];
            jointPos.Y = pmd->GetBoneChunk()[i].pos[1];
            jointPos.Z = pmd->GetBoneChunk()[i].pos[2];
        }
        else
        {
            // eꍇ͎擾Ēǉ
            int parentJoint = AnimatedMesh->getJointNumber(
                pmd->GetBoneChunk()[pmd->GetBoneChunk()[i].parent].name);
            parent = AnimatedMesh->getAllJoints()[parentJoint];

            // Irrlicht̃{[͑΍WȂ̂ŌvZ
            jointPos.X = pmd->GetBoneChunk()[i].pos[0] - pmd->GetBoneChunk()[pmd->GetBoneChunk()[i].parent].pos[0];
            jointPos.Y = pmd->GetBoneChunk()[i].pos[1] - pmd->GetBoneChunk()[pmd->GetBoneChunk()[i].parent].pos[1];
            jointPos.Z = pmd->GetBoneChunk()[i].pos[2] - pmd->GetBoneChunk()[pmd->GetBoneChunk()[i].parent].pos[2];

            joint = AnimatedMesh->createJoint(parent);
        }
        // {[̐ݒ
        joint->Name = pmd->GetBoneChunk()[i].name;
        // ΍Wݒ
        // GlobalWfinalize()ŌvZ
        joint->LocalMatrix.setTranslation(jointPos);
#if 0
        // Aj[VŎgp|WVL[쐬
        // VMDt@C쐬H
        keyP = AnimatedMesh->createPositionKey(joint);
        keyP->frame = 0;
        keyP->position = vector3df(0, 0, 0);
        keyR = AnimatedMesh->createRotationKey(joint);
        keyR->frame = 0;
        keyR->rotation.set(0, 0, 0, 1);
        keyS = AnimatedMesh->createScaleKey(joint);
        keyS->frame = 0;
        keyS->scale = vector3df(1, 1, 1);
#endif
    }

    // WCg쐬܂
}

void CPMDMeshFileLoader::createMaterialData()
{
    int materialIndexCount = 0;
    char str[256], fileName[256];

    // ꎞf[^
    u16 VertexCount, IndexCount;
    f32 *Vertices;
    u16 *Indices;
    f32 *UData, *VData;
    f32 *NVector;
    u16 *bone1, *bone2, *boneWeight;
    SMaterial *sm;

    // }eA̐JԂ
    for (int i=0; i<materialCount; i++)
    {
        sm = new SMaterial();
        // PMDt@C̃}eAeNX`擾
        sprintf_s(str, 256, "%s", pmd->GetMaterialChunk()[i].textureFileName);
        // eNX`t@Cw肳Ăꍇ
        if (strcmp(str, "") != 0)
        {
            sprintf_s(fileName, 256, "..\\..\\media\\%s", str);
            // SMaterialɃeNX`t@Cݒ
            sm->setTexture(0, VideoDriver->getTexture(fileName));
            sm->BackfaceCulling = false;
        }
        // }eA̐ݒ
        // SMaterial̃f[^0`255̐^
        sm->DiffuseColor.setRed(pmd->GetMaterialChunk()[i].diffuse.r*255.0);
        sm->DiffuseColor.setGreen(pmd->GetMaterialChunk()[i].diffuse.g*255.0);
        sm->DiffuseColor.setBlue(pmd->GetMaterialChunk()[i].diffuse.b*255.0);
        sm->DiffuseColor.setAlpha(pmd->GetMaterialChunk()[i].diffuse.a*255.0);
        sm->AmbientColor.setRed(pmd->GetMaterialChunk()[i].ambient.r*255.0);
        sm->AmbientColor.setGreen(pmd->GetMaterialChunk()[i].ambient.g*255.0);
        sm->AmbientColor.setBlue(pmd->GetMaterialChunk()[i].ambient.b*255.0);
        sm->SpecularColor.setRed(pmd->GetMaterialChunk()[i].specular.r*255.0);
        sm->SpecularColor.setGreen(pmd->GetMaterialChunk()[i].specular.g*255.0);
        sm->SpecularColor.setBlue(pmd->GetMaterialChunk()[i].specular.b*255.0);
        sm->Shininess = pmd->GetMaterialChunk()[i].shininess;

        // _̍쐬
        SSkinMeshBuffer *mesh = AnimatedMesh->createBuffer();
        // _=Index
        VertexCount = pmd->GetMaterialChunk()[i].nEdges;
        IndexCount  = pmd->GetMaterialChunk()[i].nEdges;
        // ꎞf[^̊m
        Vertices    = new irr::f32[VertexCount*3];
        Indices     = new irr::u16[IndexCount];
        UData       = new irr::f32[VertexCount];
        VData       = new irr::f32[VertexCount];
        NVector     = new irr::f32[VertexCount*3];
        bone1       = new irr::u16[VertexCount];
        bone2       = new irr::u16[VertexCount];
        boneWeight  = new irr::u16[VertexCount];

        for (int j=0; j<IndexCount; j++)
        {
            int index = materialIndexCount + j;
            // Index
            Indices[j] = pmd->GetIndexChunk()[index];
            // _W
            Vertices[3*j+0] = pmd->GetVertexChunk()[Indices[j]].x;
            Vertices[3*j+1] = pmd->GetVertexChunk()[Indices[j]].y;
            Vertices[3*j+2] = pmd->GetVertexChunk()[Indices[j]].z;
            // UV
            UData[j] = pmd->GetVertexChunk()[Indices[j]].tx;
            VData[j] = pmd->GetVertexChunk()[Indices[j]].ty;
            // @
            NVector[3*j+0] = pmd->GetVertexChunk()[Indices[j]].nx;
            NVector[3*j+1] = pmd->GetVertexChunk()[Indices[j]].ny;
            NVector[3*j+2] = pmd->GetVertexChunk()[Indices[j]].nz;
            // {[
            bone1[j] = pmd->GetVertexChunk()[Indices[j]].b1;
            bone2[j] = pmd->GetVertexChunk()[Indices[j]].b2;
            boneWeight[j] = pmd->GetVertexChunk()[Indices[j]].bw;
            // Index̏C
            Indices[j] = j;
            // EFCg̐ݒ
            ISkinnedMesh::SWeight weight1, weight2;
            // iԖڂ̃bVjԖڂ̒_
            weight1.buffer_id = i;
            weight1.vertex_id = j;
            // d݂0`1̋
            weight1.strength = (f32)boneWeight[j]/100.0;
            // iԖڂ̃bVjԖڂ̒_
            weight2.buffer_id = i;
            weight2.vertex_id = j;
            // d݂0`1̋
            weight2.strength = (f32)(100-boneWeight[j])/100.0;
            // 炩ߍ쐬ĂJointɏd݂ǉ
            AnimatedMesh->getAllJoints()[bone1[j]]->Weights.push_back(weight1);
            AnimatedMesh->getAllJoints()[bone2[j]]->Weights.push_back(weight2);
        }
        // ̃}eAIndex
        materialIndexCount += IndexCount;

        // SSkinMeshBuffer̍쐬
        printf("Vertex=%d, Index=%d\n", VertexCount, IndexCount);
        for (int k=0; k<IndexCount; k++)
        {
            video::S3DVertex temp_vertex;
            temp_vertex.Pos.X = (f32)Vertices[3*k];
            temp_vertex.Pos.Y = (f32)Vertices[3*k+1];
            temp_vertex.Pos.Z = (f32)Vertices[3*k+2];
            // SSkinMeshBufferɒ_ǉ
            mesh->Vertices_Standard.push_back(temp_vertex);
            // SSkinMeshBufferindexǉ
            mesh->Indices.push_back(Indices[k]);
        }
        for (int k=0; k<mesh->Vertices_Standard.size(); k++)
        {
            // @̐ݒ^PʃxNg
            mesh->Vertices_Standard[k].Normal.set(
                NVector[3*k], NVector[3*k+1], NVector[3*k+2]);
            mesh->Vertices_Standard[k].Normal.normalize();
        }
        // _eNX`^CvɕύX
        mesh->MoveTo_2TCoords();
        for (int k=0; k<mesh->Vertices_2TCoords.size(); k++)
        {
            // _̃eNX`Wݒ
            mesh->Vertices_2TCoords[k].TCoords.X = UData[k];
            mesh->Vertices_2TCoords[k].TCoords.Y = VData[k];
        }
        // SSkinMeshBuffer1MaterialĂȂH
        mesh->Material = *sm;
        // oEfBO{bNX̍ČvZ
        mesh->recalculateBoundingBox();
        printf("vc=%d, standard=%d\n", mesh->getVertexCount(),
            mesh->Vertices_Standard.size());

        // ꎞf[^̍폜
        delete sm;
        delete [] Vertices;
        delete [] Indices;
        delete [] UData;
        delete [] VData;
        delete [] NVector;
        delete [] bone1;
        delete [] bone2;
        delete [] boneWeight;
    }
}
