#include <irrlicht.h>
#include "common.h"
#include "CPMDMeshFileLoader.h"

using namespace irr;
using namespace core;
using namespace scene;
using namespace video;
using namespace io;
using namespace gui;

CPMDMeshFileLoader *pmdFileLoader;

static float angle = 0.0;

#pragma comment(lib, "Irrlicht.lib")

// {[̕`
void drawJoints(IVideoDriver *driver, ISkinnedMesh::SJoint *joint)
{
    irr::f32 size = 0.1;
    // Box̍쐬
    vector3df minPos = joint->GlobalMatrix.getTranslation() - vector3df(-size, -size, -size);
    vector3df maxPos = joint->GlobalMatrix.getTranslation() - vector3df( size,  size,  size);
    aabbox3df box(minPos, maxPos);
    // Box̕`
    driver->draw3DBox(box, SColor(255,0,0,255));
    // {[̐JԂ
    for (int i=0; i<joint->Children.size(); i++)
    {
        // e{[q{[֐
        driver->draw3DLine(joint->GlobalMatrix.getTranslation(),
            joint->Children[i]->GlobalMatrix.getTranslation());
        // ċA
        drawJoints(driver, joint->Children[i]);
    }
}

// {[gpeXg
void update(IAnimatedMeshSceneNode* anode)
{
    // {[V[m[h{[擾
    IBoneSceneNode *bone = anode->getJointNode("");
    if (bone == NULL)
    {
        printf("{[擾ł܂łB\n");
    }
    else
    {
        // {[]
        vector3df rotation;
        angle += 0.5;
        if (angle >= 360)
        {
            angle = 0;
        }
        rotation.Z = 10.0*sin(ToRadian(angle));
        rotation.X = 0;
        rotation.Y = 0;
        bone->setRotation(rotation);
    }
}

int main(int argc, char argv[])
{
    // RĂȂƃt@CEEE
    std::locale::global(std::locale("japanese"));

    // foCX̍쐬
    IrrlichtDevice *device =
        createDevice( video::EDT_OPENGL, dimension2d<s32>(640, 480), 16,
        false, true, false, 0);

    if (!device)
    {
        return 1;
    }

    device->setWindowCaption(L"Hello World! - Irrlicht Engine Demo");

    IVideoDriver* driver = device->getVideoDriver();
    ISceneManager* smgr = device->getSceneManager();

    // room̍쐬
    // Tutorial̂܂܂łB
    scene::IAnimatedMesh* mesh = smgr->getMesh("../../media/room.3ds");
    smgr->getMeshManipulator()->makePlanarTextureMapping(mesh->getMesh(0), 0.004f);
    scene::ISceneNode* node = 0;
    node = smgr->addAnimatedMeshSceneNode(mesh);
    node->setMaterialTexture(0, driver->getTexture("../../media/wall.jpg"));
    node->getMaterial(0).SpecularColor.set(0,0,0,0);

    mesh = smgr->addHillPlaneMesh("myHill",
        core::dimension2d<f32>(20,20),
        core::dimension2d<u32>(40,40), 0, 0,
        core::dimension2d<f32>(0,0),
        core::dimension2d<f32>(10,10));

    node = smgr->addWaterSurfaceSceneNode(mesh->getMesh(0), 3.0f, 300.0f, 30.0f);
    node->setPosition(core::vector3df(0,7,0));

    node->setMaterialTexture(0, driver->getTexture("../../media/stones.jpg"));
    node->setMaterialTexture(1, driver->getTexture("../../media/water.jpg"));
    node->setMaterialType(video::EMT_REFLECTION_2_LAYER);

    // create light
    node = smgr->addLightSceneNode(0, core::vector3df(0,0,0),
        video::SColorf(1.0f, 0.6f, 0.7f, 1.0f), 800.0f);
    scene::ISceneNodeAnimator* anim = 0;
    // ̓ύX
    anim = smgr->createFlyCircleAnimator (core::vector3df(0,150,-150),100.0f, 0.001, vector3df(0,0,1));
    node->addAnimator(anim);
    anim->drop();

    // attach billboard to light
    // Tutorial̂܂܂łB
    node = smgr->addBillboardSceneNode(node, core::dimension2d<f32>(50, 50));
    node->setMaterialFlag(video::EMF_LIGHTING, false);
    node->setMaterialType(video::EMT_TRANSPARENT_ADD_COLOR);
    node->setMaterialTexture(0, driver->getTexture("../../media/particlewhite.bmp"));

    // create a particle system
    // Tutorial̂܂܂łB
    scene::IParticleSystemSceneNode* ps =
        smgr->addParticleSystemSceneNode(false);

    scene::IParticleEmitter* em = ps->createBoxEmitter(
        core::aabbox3d<f32>(-7,0,-7,7,1,7), // emitter size
        core::vector3df(0.0f,0.06f,0.0f),   // initial direction
        80,100,                             // emit rate
        video::SColor(0,255,255,255),       // darkest color
        video::SColor(0,255,255,255),       // brightest color
        800,2000,0,                         // min and max age, angle
        core::dimension2df(10.f,10.f),         // min size
        core::dimension2df(20.f,20.f));        // max size

    ps->setEmitter(em); // this grabs the emitter
    em->drop(); // so we can drop it here without deleting it

    scene::IParticleAffector* paf = ps->createFadeOutParticleAffector();

    ps->addAffector(paf); // same goes for the affector
    paf->drop();

    ps->setPosition(core::vector3df(-70,60,40));
    ps->setScale(core::vector3df(2,2,2));
    ps->setMaterialFlag(video::EMF_LIGHTING, false);
    ps->setMaterialFlag(video::EMF_ZWRITE_ENABLE, false);
    ps->setMaterialTexture(0, driver->getTexture("../../media/fire.bmp"));
    ps->setMaterialType(video::EMT_TRANSPARENT_VERTEX_ALPHA);
    scene::IVolumeLightSceneNode * n = smgr->addVolumeLightSceneNode(0, -1,
        32,                              // Subdivisions on U axis
        32,                              // Subdivisions on V axis
        video::SColor(0, 255, 255, 255), // foot color
        video::SColor(0, 0, 0, 0));      // tail color

    if (n)
    {
        n->setScale(core::vector3df(56.0f, 56.0f, 56.0f));
        n->setPosition(core::vector3df(-120,50,40));

        // load textures for animation
        core::array<video::ITexture*> textures;
        for (s32 g=7; g > 0; --g)
        {
            core::stringc tmp;
            tmp = "../../media/portal";
            tmp += g;
            tmp += ".bmp";
            video::ITexture* t = driver->getTexture( tmp.c_str() );
            textures.push_back(t);
        }

        // create texture animator
        scene::ISceneNodeAnimator* glow = smgr->createTextureAnimator(textures, 150);

        // add the animator
        n->addAnimator(glow);

        // drop the animator because it was created with a create() function
        glow->drop();
    }

    // PMDt@C[_NX̍쐬^ǉ
    pmdFileLoader = new CPMDMeshFileLoader(smgr, driver);
    smgr->addExternalMeshLoader(pmdFileLoader);

    // PMDt@CISkinnedMeshǂݍ
    ISkinnedMesh *amesh;
    amesh = (ISkinnedMesh*)smgr->getMesh("../../media/~NVer2.pmd");

    if (amesh == NULL)
    {
        printf("Can't load PMD file!\n");
        getchar();
    }

    scene::IAnimatedMeshSceneNode* anode = 0;
    anode = smgr->addAnimatedMeshSceneNode(amesh);
    anode->setPosition(core::vector3df(30,20,0));
    anode->setMaterialFlag(video::EMF_BACK_FACE_CULLING, false);

#if 0
    // add shadow(_ƌvZɎԂ܂)
    printf("Shadow VoluměvZłBBB\n");
    anode->addShadowVolumeSceneNode();
    smgr->setShadowColor(video::SColor(150,0,0,0));
#endif

    // gƖ@xNgČvZ
    anode->setScale(core::vector3df(10,10,10));
    anode->setMaterialFlag(video::EMF_NORMALIZE_NORMALS, true);

#if 1
    // {[擾eXg
    if (anode)
    {
        anode->setJointMode(static_cast<irr::scene::E_JOINT_UPDATE_ON_RENDER>(2));
        anode->animateJoints();
        printf("jointNum=%d\n", amesh->getJointCount());
        for (int i=0; i<amesh->getJointCount(); i++)
        {
            printf("jointName[%d]=%s\n", i, amesh->getJointName(i));
        }
    }
#endif

    // J̐ݒ
    scene::ICameraSceneNode* camera = smgr->addCameraSceneNodeFPS();
    camera->setPosition(core::vector3df(-50,50,-150));

    s32 lastFPS = -1;

    while(device->run())
    {
        // `揈
        driver->beginScene(true, true, SColor(255,100,101,140));
#if 0
        drawJoints(driver, amesh->getAllJoints()[0]);
#endif
        update(anode);
        smgr->drawAll();
        driver->endScene();

        // FPS擾
        const s32 fps = driver->getFPS();

        // ^CgXV
        if (lastFPS != fps)
        {
            core::stringw str = L"Irrlicht Engine - SpecialFX example [";
            str += driver->getName();
            str += "] FPS:";
            str += fps;

            device->setWindowCaption(str.c_str());
            lastFPS = fps;
        }

        if (::GetAsyncKeyState(VK_ESCAPE))
        {
            break;
        }
    }
    device->drop();

    return 0;
}